@@ -17,6 +17,7 @@
 #include <netinet/tcp.h>
 #include <arpa/inet.h>
 #include <sys/ioctl.h>
+#include <limits.h>
 
 /* for sendfile(2) */
 #ifdef __linux__
@@ -175,12 +176,17 @@ xsendfile(int fd, int sock)
 
 /* Different sendfile(2) implementations on different platforms. :/ */
 #ifdef __linux__
-	return sendfile(sock, fd, NULL, count);
+	for (sent = 0; count > 0; count -= sent) {
+		sent = sendfile(sock, fd, 0, count);
+		if (sent < 0)
+			return -1;
+	}
 #endif
 #if defined(__FreeBSD__) || defined(__DragonFly__)
 	return sendfile(fd, sock, 0, count, NULL, NULL, 0);
 #endif
-	return -1;
+
+	return 0;
 }
 
 void *
@@ -284,14 +290,9 @@ freeindex(Indexs *i)
 void
 addelem(Elems *e, char *s)
 {
-	int slen;
-
-	slen = strlen(s) + 1;
-
 	e->num++;
 	e->e = xrealloc(e->e, sizeof(char *) * e->num);
-	e->e[e->num - 1] = xcalloc(1, slen);
-	strncpy(e->e[e->num - 1], s, slen - 1);
+	e->e[e->num - 1] = xstrdup(s);
 
 	return;
 }
@@ -408,8 +409,11 @@ scanfile(char *fname)
 }
 
 int
-printelem(int fd, Elems *el, char *addr, char *port)
+printelem(int fd, Elems *el, char *file, char *base, char *addr, char *port)
 {
+	char *path, *p, buf[PATH_MAX+1];
+	int len;
+
 	if (!strcmp(el->e[3], "server")) {
 		free(el->e[3]);
 		el->e[3] = xstrdup(addr);
@@ -418,6 +422,36 @@ printelem(int fd, Elems *el, char *addr, char *port)
 		free(el->e[4]);
 		el->e[4] = xstrdup(port);
 	}
+
+	/*
+	 * Ignore if the path is from base, if it might be some h type with
+	 * some URL and ignore various types that have different semantics but
+	 * to point to some file or directory.
+	 */
+	if (el->e[2][0] != '/'
+			&& strncmp(el->e[2], "URL:", 4)
+			&& el->e[0][0] != 'i'
+			&& el->e[0][0] != '2'
+			&& el->e[0][0] != '3'
+			&& el->e[0][0] != '8'
+			&& el->e[0][0] != 'w'
+			&& el->e[0][0] != 'T') {
+		path = file + strlen(base);
+		if ((p = strrchr(path, '/')))
+			len = p - path;
+		else
+			len = strlen(path);
+		snprintf(buf, sizeof(buf), "%s%.*s/%s", base, len, path, el->e[2]);
+
+		if ((path = realpath(buf, NULL)) &&
+				!strncmp(base, path, strlen(base))) {
+			p = path + strlen(base);
+			free(el->e[2]);
+			el->e[2] = xstrdup(p[0]? p : "/");
+		}
+		free(path);
+	}
+
 	if (dprintf(fd, "%.1s%s\t%s\t%s\t%s\r\n", el->e[0], el->e[1], el->e[2],
 			el->e[3], el->e[4]) < 0) {
 		perror("printelem: dprintf");
@@ -471,7 +505,10 @@ void
 setcgienviron(char *file, char *path, char *port, char *base, char *args,
 		char *sear, char *ohost, char *chost)
 {
-
+	/*
+	 * TODO: Clean environment from possible unsafe environment variables.
+	 *       But then it is the responsibility of the script writer.
+	 */
 	unsetenv("AUTH_TYPE");
 	unsetenv("CONTENT_LENGTH");
 	unsetenv("CONTENT_TYPE");
