@@ -40,7 +40,7 @@ Indexs *scanfile(char *fname);
 Elems *getadv(char *str);
 int pendingbytes(int sock);
 void waitforpendingbytes(int sock);
-int printelem(int fd, Elems *el, char *addr, char *port);
+int printelem(int fd, Elems *el, char *file, char *base, char *addr, char *port);
 void addindexs(Indexs *idx, Elems *el);
 void addelem(Elems *e, char *s);
 void freeindex(Indexs *i);
