@@ -23,6 +23,8 @@
 #include <grp.h>
 #include <errno.h>
 #include <arpa/inet.h>
+#include <sys/select.h>
+#include <sys/time.h>
 
 #include "ind.h"
 #include "handlr.h"
@@ -39,12 +41,13 @@ enum {
 
 int glfd = -1;
 int loglvl = 15;
-int listfd = -1;
+int *listfds = NULL;
+int nlistfds = 0;
 int revlookup = 1;
 char *logfile = NULL;
 
 char *argv0;
-char *stdbase = "/var/gopher";
+char stdbase[] = "/var/gopher";
 char *stdport = "70";
 char *indexf[] = {"/index.gph", "/index.cgi", "/index.dcgi", "/index.bin"};
 char *nocgierr = "3Sorry, execution of the token '%s' was requested, but this "
@@ -66,8 +69,8 @@ char *htredir = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
 		"  </body>\n"
 		"</html>\n";
 char *selinval ="3Happy helping ☃ here: "
-		"Sorry, your selector contains '..'. That's illegal here.\tErr"
-	    "\tlocalhost\t70\r\n.\r\n\r\n";
+		"Sorry, your selector does not start with / or contains '..'. "
+		"That's illegal here.\tErr\tlocalhost\t70\r\n.\r\n\r\n";
 
 int
 dropprivileges(struct group *gr, struct passwd *pw)
@@ -252,11 +255,12 @@ handlerequest(int sock, char *base, char *ohost, char *port, char *clienth,
 void
 sighandler(int sig)
 {
+	int i;
+
 	switch (sig) {
 	case SIGCHLD:
 		while (waitpid(-1, NULL, WNOHANG) > 0);
 		break;
-	case SIGHUP:
 	case SIGINT:
 	case SIGQUIT:
 	case SIGABRT:
@@ -266,10 +270,12 @@ sighandler(int sig)
 			close(glfd);
 			glfd = -1;
 		}
-		if (listfd >= 0) {
-			shutdown(listfd, SHUT_RDWR);
-			close(listfd);
+
+		for (i = 0; i < nlistfds; i++) {
+			shutdown(listfds[i], SHUT_RDWR);
+			close(listfds[i]);
 		}
+		free(listfds);
 		exit(0);
 		break;
 	default:
@@ -291,55 +297,88 @@ initsignals(void)
 	signal(SIGPIPE, SIG_IGN);
 }
 
-int
-getlistenfd(struct addrinfo *hints, char *bindip, char *port)
+/*
+ * TODO: Move Linux and BSD to Plan 9 socket and bind handling, so we do not
+ *       need the inconsistent return and exit on getaddrinfo.
+ */
+int *
+getlistenfd(struct addrinfo *hints, char *bindip, char *port, int *rlfdnum)
 {
 	char addstr[INET6_ADDRSTRLEN];
 	struct addrinfo *ai, *rp;
 	void *sinaddr;
-	int on, listfd;
+	int on, *listenfds, *listenfd, aierr, errno_save;
 
-	listfd = -1;
-
-	if (getaddrinfo(bindip, port, hints, &ai))
-		return -1;
-	if (ai == NULL)
-		return -1;
+	if ((aierr = getaddrinfo(bindip, port, hints, &ai)) || ai == NULL) {
+		fprintf(stderr, "getaddrinfo (%s:%s): %s\n", bindip, port,
+			gai_strerror(aierr));
+		exit(1);
+	}
 
+	*rlfdnum = 0;
+	listenfds = NULL;
 	on = 1;
 	for (rp = ai; rp != NULL; rp = rp->ai_next) {
-		listfd = socket(rp->ai_family, rp->ai_socktype,
+		listenfds = xrealloc(listenfds,
+				sizeof(*listenfds) * (++*rlfdnum));
+		listenfd = &listenfds[*rlfdnum-1];
+
+		*listenfd = socket(rp->ai_family, rp->ai_socktype,
 				rp->ai_protocol);
-		if (listfd < 0)
+		if (*listenfd < 0)
+			continue;
+		if (setsockopt(*listenfd, SOL_SOCKET, SO_REUSEADDR, &on,
+				sizeof(on)) < 0) {
+			close(*listenfd);
+			(*rlfdnum)--;
+			continue;
+		}
+
+		if (rp->ai_family == AF_INET6 && (setsockopt(*listenfd,
+				IPPROTO_IPV6, IPV6_V6ONLY, &on,
+				sizeof(on)) < 0)) {
+			close(*listenfd);
+			(*rlfdnum)--;
 			continue;
-		if (setsockopt(listfd, SOL_SOCKET, SO_REUSEADDR, &on,
-					sizeof(on)) < 0) {
-			close(listfd);
-			break;
 		}
 
 		sinaddr = (rp->ai_family == AF_INET) ?
 		          (void *)&((struct sockaddr_in *)rp->ai_addr)->sin_addr :
 		          (void *)&((struct sockaddr_in6 *)rp->ai_addr)->sin6_addr;
 
-		if (bind(listfd, rp->ai_addr, rp->ai_addrlen) == 0) {
+		if (bind(*listenfd, rp->ai_addr, rp->ai_addrlen) == 0) {
 			if (loglvl & CONN && inet_ntop(rp->ai_family, sinaddr,
-			    addstr, sizeof(addstr))) {
+					addstr, sizeof(addstr))) {
+				/* Do not revlookup here. */
+				on = revlookup;
+				revlookup = 0;
 				logentry(addstr, port, "-", "listening");
+				revlookup = on;
 			}
-			break;
+			continue;
 		}
-		close(listfd);
-		if (loglvl & CONN && inet_ntop(rp->ai_family, sinaddr,
-		    addstr, sizeof(addstr))) {
-			logentry(addstr, port, "-", "could not bind");
-		}
-	}
-	if (rp == NULL)
-		return -1;
-	freeaddrinfo(ai);
 
-	return listfd;
+		/* Save errno, because fprintf in logentry overwrites it. */
+		errno_save = errno;
+		close(*listenfd);
+		(*rlfdnum)--;
+		if (loglvl & CONN && inet_ntop(rp->ai_family, sinaddr,
+				addstr, sizeof(addstr))) {
+			/* Do not revlookup here. */
+			on = revlookup;
+			revlookup = 0;
+			logentry(addstr, port, "-", "could not bind");
+			revlookup = on;
+		}
+		errno = errno_save;
+	}
+	freeaddrinfo(ai);
+	if (*rlfdnum < 1) {
+		free(listenfds);
+		return NULL;
+	}
+
+	return listenfds;
 }
 
 void
@@ -347,7 +386,7 @@ usage(void)
 {
 	dprintf(2, "usage: %s [-46cden] [-l logfile] "
 	           "[-v loglvl] [-b base] [-p port] [-o sport] "
-	           "[-u user] [-g group] [-h host] [-i IP]\n",
+	           "[-u user] [-g group] [-h host] [-i interface ...]\n",
 		   argv0);
 	exit(1);
 }
@@ -358,11 +397,13 @@ main(int argc, char *argv[])
 	struct addrinfo hints;
 	struct sockaddr_storage clt;
 	socklen_t cltlen;
-	int sock, dofork, v4, v6, usechroot, nocgi;
+	int sock, dofork, inetf, usechroot, nocgi, errno_save, nbindips, i, j,
+	    nlfdret, *lfdret, listfd, maxlfd;
 	char *port, *base, clienth[NI_MAXHOST], clientp[NI_MAXSERV];
-	char *user, *group, *bindip, *ohost, *sport;
+	char *user, *group, **bindips, *ohost, *sport, *p;
 	struct passwd *us;
 	struct group *gr;
+	fd_set rfd;
 
 	base = stdbase;
 	port = stdport;
@@ -371,20 +412,20 @@ main(int argc, char *argv[])
 	group = NULL;
 	us = NULL;
 	gr = NULL;
-	bindip = NULL;
+	bindips = NULL;
+	nbindips = 0;
 	ohost = NULL;
-	sport = port;
-	v4 = 1;
-	v6 = 1;
+	sport = NULL;
+	inetf = AF_UNSPEC;
 	usechroot = 0;
 	nocgi = 0;
 
 	ARGBEGIN {
 	case '4':
-		v6 = 0;
+		inetf = AF_INET;
 		break;
 	case '6':
-		v4 = 0;
+		inetf = AF_INET6;
 		break;
 	case 'b':
 		base = EARGF(usage());
@@ -394,6 +435,8 @@ main(int argc, char *argv[])
 		break;
 	case 'p':
 		port = EARGF(usage());
+		if (sport == NULL)
+			sport = port;
 		break;
 	case 'l':
 		logfile = EARGF(usage());
@@ -414,7 +457,8 @@ main(int argc, char *argv[])
 		group = EARGF(usage());
 		break;
 	case 'i':
-		bindip = EARGF(usage());
+		bindips = xrealloc(bindips, sizeof(*bindips) * (++nbindips));
+		bindips[nbindips-1] = EARGF(usage());
 		break;
 	case 'h':
 		ohost = EARGF(usage());
@@ -429,12 +473,16 @@ main(int argc, char *argv[])
 		usage();
 	} ARGEND;
 
+	if (sport == NULL)
+		sport = port;
+
 	if (argc != 0)
 		usage();
 
 	if (ohost == NULL) {
-		ohost = xcalloc(1, 513);
-		if (gethostname(ohost, 512) < 0) {
+		/* Do not use HOST_NAME_MAX, it is not defined on NetBSD. */
+		ohost = xcalloc(1, 256+1);
+		if (gethostname(ohost, 256) < 0) {
 			perror("gethostname");
 			free(ohost);
 			return 1;
@@ -444,15 +492,25 @@ main(int argc, char *argv[])
 	}
 
 	if (group != NULL) {
+		errno = 0;
 		if ((gr = getgrnam(group)) == NULL) {
-			perror("no such group");
+			if (errno == 0) {
+				fprintf(stderr, "no such group '%s'\n", group);
+			} else {
+				perror("getgrnam");
+			}
 			return 1;
 		}
 	}
 
 	if (user != NULL) {
+		errno = 0;
 		if ((us = getpwnam(user)) == NULL) {
-			perror("no such user");
+			if (errno == 0) {
+				fprintf(stderr, "no such user '%s'\n", user);
+			} else {
+				perror("getpwnam");
+			}
 			return 1;
 		}
 	}
@@ -479,39 +537,51 @@ main(int argc, char *argv[])
 		glfd = 1;
 	}
 
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_flags = AI_PASSIVE;
-	hints.ai_socktype = SOCK_STREAM;
-	if (bindip)
-		hints.ai_flags |= AI_CANONNAME;
-
-	listfd = -1;
-	if (v6) {
-		hints.ai_family = PF_INET6;
-		listfd = getlistenfd(&hints, bindip, port);
-		if (!v4 && listfd < 0) {
-			perror("getlistenfd6");
-			return 1;
+	if (bindips == NULL) {
+		if (inetf == AF_INET || inetf == AF_UNSPEC) {
+			bindips = xrealloc(bindips, sizeof(*bindips) * (++nbindips));
+			bindips[nbindips-1] = "0.0.0.0";
+		}
+		if (inetf == AF_INET6 || inetf == AF_UNSPEC) {
+			bindips = xrealloc(bindips, sizeof(*bindips) * (++nbindips));
+			bindips[nbindips-1] = "::";
 		}
 	}
-	if (v4 && listfd < 0) {
-		hints.ai_family = PF_INET;
-		listfd = getlistenfd(&hints, bindip, port);
-		if (listfd < 0) {
-			perror("getlistenfd4");
-			return 1;
-		}
-	}
-	if (listfd < 0) {
-		perror("You did not specify a TCP port.");
-		return 1;
-	}
 
-	if (listen(listfd, 255)) {
-		perror("listen");
-		close(listfd);
-		return 1;
+	for (i = 0; i < nbindips; i++) {
+		memset(&hints, 0, sizeof(hints));
+		hints.ai_family = inetf;
+		hints.ai_flags = AI_PASSIVE;
+		hints.ai_socktype = SOCK_STREAM;
+		if (bindips[i])
+			hints.ai_flags |= AI_CANONNAME;
+
+		nlfdret = 0;
+		lfdret = getlistenfd(&hints, bindips[i], port, &nlfdret);
+		if (nlfdret < 1) {
+			errno_save = errno;
+			fprintf(stderr, "Unable to get a binding socket for "
+					"%s:%s\n", bindips[i], port);
+			errno = errno_save;
+			perror("getlistenfd");
+		}
+
+		for (j = 0; j < nlfdret; j++) {
+			if (listen(lfdret[j], 255) < 0) {
+				perror("listen");
+				close(lfdret[j]);
+				continue;
+			}
+			listfds = xrealloc(listfds,
+					sizeof(*listfds) * ++nlistfds);
+			listfds[nlistfds-1] = lfdret[j];
+		}
+		free(lfdret);
 	}
+	free(bindips);
+
+	if (nlistfds < 1)
+		return 1;
 
 	if (usechroot) {
 		if (chdir(base) < 0) {
@@ -528,21 +598,27 @@ main(int argc, char *argv[])
 		return 1;
 	}
 
+	/* strip / at the end, except if it is "/" */
+	for (p = base + strlen(base); p > base + 1 && p[-1] == '/'; --p)
+		p[-1] = '\0';
+
 	if (dropprivileges(gr, us) < 0) {
 		perror("dropprivileges");
-		close(listfd);
+
+		for (i = 0; i < nlistfds; i++) {
+			shutdown(listfds[i], SHUT_RDWR);
+			close(listfds[i]);
+		}
+		free(listfds);
 		return 1;
 	}
 
 	initsignals();
 
-	cltlen = sizeof(clt);
-
 #ifdef __OpenBSD__
-	char *promises = (char[35]){};
-	snprintf(promises, 35, "rpath inet stdio proc exec %s %s",
-		revlookup ? "dns"  : "",
-		dofork    ? "tty"  : "");
+	char promises[31]; /* check the size needed in the fork too */
+	snprintf(promises, sizeof(promises), "rpath inet stdio proc exec %s",
+	         revlookup ? "dns" : "");
 	if (pledge(promises, NULL) == -1) {
 		perror("pledge");
 		exit(1);
@@ -550,6 +626,32 @@ main(int argc, char *argv[])
 #endif /* __OpenBSD__ */
 
 	while (1) {
+		FD_ZERO(&rfd);
+		maxlfd = 0;
+		for (i = 0; i < nlistfds; i++) {
+			FD_SET(listfds[i], &rfd);
+			if (listfds[i] > maxlfd)
+				maxlfd = listfds[i];
+		}
+
+		if (pselect(maxlfd+1, &rfd, NULL, NULL, NULL, NULL) < 0) {
+			if (errno == EINTR)
+				continue;
+			perror("pselect");
+			break;
+		}
+
+		listfd = -1;
+		for (i = 0; i < nlistfds; i++) {
+			if (FD_ISSET(listfds[i], &rfd)) {
+				listfd = listfds[i];
+				break;
+			}
+		}
+		if (listfd < 0)
+			continue;
+
+		cltlen = sizeof(clt);
 		sock = accept(listfd, (struct sockaddr *)&clt, &cltlen);
 		if (sock < 0) {
 			switch (errno) {
@@ -563,9 +665,11 @@ main(int argc, char *argv[])
 			}
 		}
 
-		getnameinfo((struct sockaddr *)&clt, cltlen, clienth,
+		if (getnameinfo((struct sockaddr *)&clt, cltlen, clienth,
 				sizeof(clienth), clientp, sizeof(clientp),
-				NI_NUMERICHOST|NI_NUMERICSERV);
+				NI_NUMERICHOST|NI_NUMERICSERV)) {
+			clienth[0] = clientp[0] = '\0';
+		}
 
 		if (!strncmp(clienth, "::ffff:", 7))
 			memmove(clienth, clienth+7, strlen(clienth)-6);
@@ -588,8 +692,10 @@ main(int argc, char *argv[])
 			signal(SIGALRM, SIG_DFL);
 
 #ifdef __OpenBSD__
-			promises = nocgi ? "rpath inet stdio" :
-			                   "rpath inet stdio proc exec";
+			snprintf(promises, sizeof(promises),
+			         "rpath inet stdio %s %s",
+			         nocgi     ? ""    : "proc exec",
+			         revlookup ? "dns" : "");
 			if (pledge(promises, NULL) == -1) {
 				perror("pledge");
 				exit(1);
@@ -614,14 +720,18 @@ main(int argc, char *argv[])
 		close(sock);
 	}
 
-	shutdown(listfd, SHUT_RDWR);
-	close(listfd);
 	if (logfile != NULL && glfd != -1) {
 		close(glfd);
 		glfd = -1;
 	}
 	free(ohost);
 
+	for (i = 0; i < nlistfds; i++) {
+		shutdown(listfds[i], SHUT_RDWR);
+		close(listfds[i]);
+	}
+	free(listfds);
+
 	return 0;
 }
 
